﻿using WingRpcSDK.Common.Loader;
using WingRpcSDK.JsonRpc.Driver;
using WingRpcSDK.Log;
using JsonRpcLite.InProcess;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;

namespace WingCloudServer
{
    internal abstract class BaseServiceLoader
    {
        internal static object RestartLocker = new object();
        /// <summary>
        /// RpcInProcessClient
        /// </summary>
        internal static InProcessRpcClient RpcInProcessClient { get; }

        /// <summary>
        /// RpcInProcessService
        /// </summary>
        internal static RpcServer RpcInProcessService { get; }

        static BaseServiceLoader()
        {
            var inProcessEngine = new JsonRpcInProcessEngine();
            RpcInProcessService = new RpcServer(inProcessEngine);
            RpcInProcessClient = InProcessRpcClient.CreateClient(inProcessEngine);
            RpcInProcessService.Start();
        }

        /// <summary>
        /// Create service
        /// </summary>
        /// <typeparam name="T">Service convert type</typeparam>
        /// <param name="serviceType">Create source type</param>
        /// <returns>return need type created instance</returns>
        private T CreateLoader<T>(Type serviceType)
        {
            var service = (T)Activator.CreateInstance(serviceType);
            return service;
        }

        /// <summary>
        /// Register service
        /// </summary>
        /// <typeparam name="T">Service type</typeparam>
        /// <param name="service">Service instance</param>
        /// <param name="rpcService">Run rpc service instance</param>
        /// <param name="interfaceType">The service implement from interface type</param>
        internal void CallRegister(RpcServer rpcServer,Type serviceType, object service)
        {
            var registerMethod = typeof(RpcServer).GetMethod("RegisterService");
            registerMethod.Invoke(rpcServer, new object[] { serviceType, service });
        }

        /// <summary>
        /// Get the service base type,used to create and register service  
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="filePath"></param>
        /// <returns></returns>
        internal IEnumerable<Type> GetLoaders<T>(string filePath)
        {
            var assemblyBytes = File.ReadAllBytes(filePath);
            var assembly = Assembly.Load(assemblyBytes);
            var allTypes = assembly.GetTypes();
            return allTypes.Where(t => t.GetInterface(typeof(T).Name)!=null);
        }

        /// <summary>
        /// Load service dll
        /// </summary>
        internal virtual void LoadDlls(string path)
        {
            var inProcessDll = Directory.GetFiles(path,"*.dll");
            var firstLoad = inProcessDll.Where(f => new List<string> { "WingMongoDBModule.dll" }.Contains(Path.GetFileName(f))).ToList();
            var afterLoad = inProcessDll.Except(firstLoad);
            firstLoad.AddRange(afterLoad);
            foreach (var file in firstLoad)
            {
                var serviceLoaders = GetLoaders<IServiceLoader>(file);
                foreach (var serviceLoader in serviceLoaders)
                {
                    var loader = CreateLoader<IServiceLoader>(serviceLoader);
                    var serviceType = loader.GetServiceType();
                    var service = loader.CreateService();
                    RegisterService(serviceType, service);
                }
            }
        }

        /// <summary>
        /// Start service
        /// </summary>
        internal abstract void StartService();

        /// <summary>
        /// Stop service
        /// </summary>
        internal abstract void StopService();

        /// <summary>
        /// Register servcie
        /// </summary>
        /// <param name="serviceType">service type</param>
        /// <param name="service">instance</param>
        internal abstract void RegisterService(Type serviceType, object service);
    }
}
